
module main;

import std.stdio;
import std.traits;
import std.typetuple;

import metaclass;
static import reflect;
import std.variant;

class Bar : Foo {
    override void foo( ... ) { writeln( "hello! ", _arguments.length ); }
}

class Foo {

    void foo( ... ) { writeln( "Hello!" ); }
    
    void hello( string greet ) { writeln( "Hello! ", greet ); }

    static int bar() { writeln( "Hello Static!" ); return 1; }

    double bar( double i ) { writeln( "Hello Void! " ); return 1; }
    int bar( int i ) { writeln( "Hello Int! ", i ); return i; }
    final int bar( int i, ref int j ) { writeln( "Hello Int 2x! ", i, j ); j = i * j; return j; }
    
    int x, y, z, w;
}

int main()
{
    Foo bar = new Bar();
    auto foo = new Foo();
    
    // Testing Method
    
    auto mm = Method!( Foo, "bar" )();
    
    // Prints all overloads. Overloads have toString overloaded to print the method signature.
    foreach( i, overload; Method!( Foo, "foo" ).overloadTuple ) {
        writeln( overload );
    }
    foreach( i, overload; Method!( Foo, "bar" ).overloadTuple ) {
        writeln( overload );
    }
    
    // Using toString, prints all overloads in an array style.
    writeln( Method!( Foo, "bar" )() );

    // Testing compile time reflection -----
    
    writeln( "\nTesting compile time reflection using Class\n" );
    {
        auto cls = new Class!( Foo )();
        cls.invoke!"bar"( foo, cast(short)2, 2 );
        cls.invokeStatic!("bar", int)();
        writeln( "rval ", cls.invoke!("bar",int)( foo, cast(short)2, 2 ) );

        cls.invokeStatic!"bar"();
        cls.invoke!"foo"(foo, 1);
    }
        
    writeln( "\nTesting compile time reflection using MetaObject\n" );   
    { 
        auto mo = new MetaObject!( Foo )( bar );
        mo.invoke!"bar"( 1 );
        
        mo.invoke!"bar"( cast(short)2, 2 );
        mo.invoke!("bar", int)();
        writeln( "rval ", mo.invoke!("bar",int)( cast(short)2, 2 ) );

        mo.invoke!"bar"();
        mo.invoke!"bar"(1);
        
        mo.invoke!"foo"();
    }
    
    // Testing run time reflection -----
    
    writeln( "\nTesting run time reflection using MetaObject\n" );
    {    
        auto cls = new reflect.Class!Foo();

        string methodName = "hello";
        string message = "reflection rocks!";
        cls.invoke( foo, methodName, message );

        cls.invoke( foo, "bar", 1 );
        
        cls.invoke( foo, "bar", 2, 2 );
        
        // cls.invoke("bar");
        
        writeln("rval ", cls.invoke!int( foo, "bar", 2, 2 ) );

        cls.invoke( foo, "bar" );
        cls.invoke( foo, "bar", 1 );
    }   

    writeln( "\nTesting run time reflection using MetaObject\n" );
    {    
        auto dmo = new reflect.MetaObject!Foo( foo );
        
        dmo.invoke( "bar", 1 );
        
        string methodName = "hello";
        string message = "reflection rocks!";
        dmo.invoke( methodName, message );
        
        dmo.invoke( "bar", 2, 2 );
        
        // dmo.invoke("bar");
        
        writeln( "rval ", dmo.invoke!int("bar", 2, 2 ) );

        dmo.invoke( "bar" );
        dmo.invoke( "bar", 1 );
    }   
 
    writeln( "\nTesting run time reflection using VariantInvokable interface.\n" ); 
    {   
        reflect.VariantInvokable dmo = new reflect.MetaObject!Foo( foo );
        
        dmo.variantInvoke( "bar", Variant(1) );
        
        string methodName = "hello";
        string message = "reflection rocks!";
        dmo.variantInvoke( methodName, Variant(message) );
        
        dmo.variantInvoke( "bar", variantArray(2, 2) );
        
        dmo.variantInvoke( "bar" );
        
        writeln( "rval ", dmo.variantInvoke("bar", variantArray(2, 2) ) );

        dmo.variantInvoke( "bar" );
        dmo.variantInvoke( "bar", Variant(1) );
    }
    
    return 0;
}
