
static import metaclass;

import std.stdarg;
import std.variant;
import std.traits;

interface VariantInvokable {
    Variant variantInvoke( in string methodName, Variant[] args... );
}

class MetaObject( TType ) : VariantInvokable {

    this( TType object ) {
        this.object = object;
    }

    alias .Class!(TType) Class;
    
    TReturnType invoke( TReturnType = void )( in string methodName, ... ) {
        static if( is(TReturnType  == void) )
            reflection.variadicInvoke( object, methodName, _arguments, _argptr );
        else
            return reflection.variadicInvoke!TReturnType( object, methodName, _arguments, _argptr );
    }
    
    override Variant variantInvoke( in string methodName, Variant[] args... ) {
        return reflection.variantInvoke( object, methodName, args );
    }
    
    TType object;    
    static const Class reflection;
    
    static this () {
        reflection = new Class();
    }
}

interface VariantClassInvokable( TType ) {
    const Variant variantInvoke( TType object, in string methodName, Variant[] args... );
}

class Class( TType ) : VariantClassInvokable!(TType) {

    alias metaclass.Class!(TType) MetaClass;

    const TReturnType invoke( TReturnType = void )( TType object, in string methodName, ... ) { 
        static if( is(TReturnType  == void) )
            variadicInvoke( object, methodName, _arguments, _argptr );
        else
            return variadicInvoke!TReturnType( object, methodName, _arguments, _argptr );
    }
    
    
    // TODO : These next two functions are *very* similar, but I could not find a decent way to
    //        modularize it. I'm thinking about using mixins. That would really suck.
    override const Variant variantInvoke( TType object, in string methodName, Variant[] args... ) {
        foreach( i, method; MetaClass.methodTuple ) {
            if( method.name == methodName ) {
                foreach( overload; method.overloadTuple ) {
                
                    if( overload.parameterTypeTuple.length != args.length )
                        continue;
                        
                    overload.parameterTypeTuple parameters;
                    
                    bool match = true;
                    foreach( j, parameterType; overload.parameterTypeTuple ) {
                        // TODO : Create a way to do implicit conversion where applyable.
                        //        Equality comparison is too strict
                        if( args[j].type != typeid(parameterType) )
                        { 
                            match = false; 
                            break; // Parameter mismatch
                        }
                            
                        parameters[j] = args[j].get!parameterType();
                    }
                    if( match ) {
                        static if( is(TReturnType == void) || is(overload.returnType == void) ) {
                            MetaClass.invoke!( method.name, overload.returnType, overload.parameterTypeTuple )( object, parameters );
                            return Variant( null );
                        }
                        else static if( Variant.allowed!(overload.returnType) ) {
                            return Variant( MetaClass.invoke!( method.name, overload.returnType, overload.parameterTypeTuple )( object, parameters ) );
                        }
                        else
                            throw new Exception( "Return type mismatch. Could not convert " ~ typeid(overload.returnType).toString() ~ " to Variant" );
                    }
                }
                
                break; // Found method, but no overload match.
            }
        }
        
        throw new Exception( "Could not find a method match!" );  
    }
    
    const TReturnType variadicInvoke(TReturnType = void)( TType object, in string methodName, TypeInfo[] _arguments, void* _argptr ) {
        foreach( i, method; MetaClass.methodTuple ) {
            if( method.name == methodName ) {
                foreach( overload; method.overloadTuple ) {
                
                    if( overload.parameterTypeTuple.length != _arguments.length )
                        continue;

                    auto args = _argptr;
                        
                    overload.parameterTypeTuple parameters;
                    
                    bool match = true;
                    foreach( j, parameterType; overload.parameterTypeTuple ) {
                        // TODO : Create a way to do implicit conversion where applyable.
                        //        Equality comparison is too strict
                        if( _arguments[j] != typeid(parameterType) )
                        { 
                            match = false; 
                            break; // Parameter mismatch
                        }
                            
                        parameters[j] = va_arg!parameterType( args );
                    }
                    if( match ) {
                        static if( is(TReturnType == void) ) {
                            MetaClass.invoke!( method.name, overload.returnType, overload.parameterTypeTuple )( object, parameters );
                            return;
                        }
                        else static if( isImplicitlyConvertible!(overload.returnType, TReturnType) ) {
                            return MetaClass.invoke!( method.name, overload.returnType, overload.parameterTypeTuple )( object, parameters );
                        }
                        else
                            throw new Exception( "Return type mismatch. Expected: " ~ typeid(overload.returnType).toString() ~ " Got: " ~ typeid(TReturnType).toString() );
                    }
                }
                
                break; // Found method, but no overload match.
            }
        }
        
        throw new Exception( "Could not find a method match!" );    
    }
}
